"""Simple container of utils."""

import numpy as np


def extract_numpy_data(arr: np.ndarray):
    """Extract data from numpy array."""
    if arr.dtype.kind == 'S':  # String
        return arr.tobytes()
    elif len(arr.data.shape) == 0:  # Single value
        return arr.item()
    elif arr.data.shape[0] > 1:  # List of vals
        return arr.tolist()
    elif arr.data.shape[0] == 1:  # Single value in brackets
        return arr[0]
    else:
        raise TypeError('Unsure how to extract from this numpy format.')



"""gxsm filename parsing.

This file contains the logic to parse a gxsm filename into its component parts.
The included class allows easy access to these components (explained in the
class documentation).
"""

from dataclasses import dataclass
import os

GXSM_FILENAME_ATTRIB_SEPARATOR = '-'
GXSM_FORWARD_SCAN_DIR = 'Xp'
GXSM_BACKWARD_SCAN_DIR = 'Xm'

@dataclass
class GxsmFileAttribs:
    """Class to hold the filename attributes, decoded.

    A gxsm filename is of the form:
        $file_base$[-M-]-$scan_direction$-$channel$.nc

    where:
        - $file_base$ is the user-provided filename;
        - '-M-' indicates if it is the 'main file' of the recording (impling it
            contains the most metadata).
        - $scan_direction$ indicates the direction of scan while this data was
            recorded: Xp == --> (presumably forward), Xm == <-- (presumably
            backward).
        - $channel$ indicates the data channel. It will either be 'Topo' for
            topography, or 'ADC#[$extra_stuff$]', indicating the explicit ADC
            channel it is collecting (these are the input ports of the
            collection module).
        - brackets ([]) indicates an optional parameter.
    """

    file_base: str
    channel: str
    scan_direction: str
    is_main_file: bool


def get_unique_channel_name(channel: str, scan_direction: str) -> str:
    """Provide a unique channel name.

    Since a file can contain data from the same channel twice
    (once per scan direction), we provide this helper method.
    """
    return channel + GXSM_FILENAME_ATTRIB_SEPARATOR + \
        scan_direction


def parse_gxsm_filename(filename: str) -> GxsmFileAttribs:
    """Parse the gxsm filename into components.

    Args:
        filename: string to parse

    Returns:
        A GxsmFilename instance, indicating the filename attributes.
    """
    basename = os.path.basename(os.path.splitext(filename)[0])
    substrs = basename.split(GXSM_FILENAME_ATTRIB_SEPARATOR)
    is_main_file = len(substrs) == 4
    if is_main_file:
        del substrs[1]  # Remove "-M-" from substrs
    return GxsmFileAttribs(substrs[0], substrs[2], substrs[1],
                           is_main_file)



"""Pre-processing options to be performed on a gxsm file.

This file contains methods that can be performed on gxsm data
in order to help convert it to more 'understandable' data
(primarily, converting to physical units data).

We call it pre-processing because it will be bundled in a pre-
processing callable if called using xarray.open_mfdataset().

If using on it's own, you can simply feed an xarray.Dataset
to be corrected.

Note: these methods assume the provided dataset has been validated
as corresponding to a gxsm file!
"""

import xarray
from . import filename as fn
from . import utils
from . import channel_config  as cc

# Lists the expected gxsm data variables and coords that are *not* metadata and
# should be kept.
GXSM_DATA_VAR = 'FloatField'
GXSM_DATA_DIFFERENTIAL = 'dz'

GXSM_KEPT_COORDS = ['dimx', 'dimy']
GXSM_KEPT_DATA_VARS = [GXSM_DATA_VAR, GXSM_DATA_DIFFERENTIAL]

# For it to be a proper gxsm file, we expect this attr key and val
GXSM_FORMAT_CHECK = ('Creator', 'gxsm')


def preprocess(ds: xarray.Dataset,
               use_physical_units: bool,
               allow_convert_from_metadata: bool,
               simplify_metadata: bool,
               channels_config_dict: dict | None
               ) -> xarray.Dataset:
    """Convert floatfield and (optionally) simplify metadata.

    This is the top-level handler for preprocessing the gxsm dataset into our
    desired format.

    Args:
        ds: the Dataset instance we are to convert, assumed to be from a gxsm
            data file.
        use_physical_units: whether or not to record the data in physical
            units. If true, we require 'conversion_factor'  and 'units'
            to exist (see optional exception below). Else, 'units' will be
            'raw' and 'conversion_factor' '1.0'.
        allow_convert_from_metadata: for some channels, there are hardcoded
            gxsm metadata attributes that contain the V-to-units conversion.
            If this attribute is true, we will use the metadata conversion
            as a fallback (i.e. if the config does not contain it).
        simplify_metadata: whether or not to convert all metadata variables
            to attributes.
        channels_config_dict: a dict containing gxsm channel configuration data
            (including the V-to-x unit conversion for each channel).
    """
    if not is_gxsm_file(ds):
        raise TypeError('The provided file does not appear to be a gxsm file!')

    # Note: Could also use ds['basename'].data.tobytes() [gxsm-specific]
    # (but this is the filepath on the device it was first recorded).
    filename = ds.encoding['source']
    gxsm_file_attribs = fn.parse_gxsm_filename(filename)
    channel_config = cc.CreateGxsmChannelConfig(channels_config_dict, ds,
                                                gxsm_file_attribs,
                                                use_physical_units,
                                                allow_convert_from_metadata)
    ds = clean_floatfield(ds)
    ds = clean_kept_coords(ds)
    ds = convert_floatfield(ds, channel_config)
    if simplify_metadata:
        ds = clean_up_metadata(ds, [channel_config.name])
    return ds


def clean_floatfield(ds: xarray.Dataset) -> xarray.Dataset:
    """Remove spurious dimensions from FloatField array.

    The main data array, 'FloatField', has 2 spurious dimensions: 'time' and
    'value'. This makes the array of the form [1,1,dimy,dimy,...]. This is
    unnecessary and makes it hard to remove unnecessary meatadata/coords
    later.

    Thus, this helper just explicitly removes those first 2 dimensions!

    Args:
        ds: input xarray.Dataset, presumably with original 'FloatField' da.

    Returns:
        output xarray.Dataset, with the first two dimensions of the data
        removed.
    """
    da = xarray.DataArray(
        data=ds[GXSM_DATA_VAR].data[0, 0, ...],
        dims=['dimy', 'dimx'],
        coords={'dimy': ds.dimy, 'dimx': ds.dimx})

    ds[GXSM_DATA_VAR] = da
    return ds


def convert_floatfield(ds: xarray.Dataset, channel_config: cc.GxsmChannelConfig
                       ) -> xarray.Dataset:
    """Convert gxsm file 'FloatField' variable to 'raw' or 'physical' data.

    gxsm stores its recorded data in a 'FloatField' attribute, containing
    raw DAC counter data. In order to convert it to raw units, we must
    perform the following:
        data['raw'] = data['FloatField'] * data['dz']

    To convert it to physical units:
        data['physical'] = data['FloatField'] * data['dz'] * V_to_x_conversion

    where:
    - data['dz'] is the differential in 'z', correlating a DAC counter to
        a physical unit *within gxsm's understanding of the world*. Since the
        DAC data is received from one of the hardware device input channels
        (ADC#), this 'dz' is used to convert from the DC received voltage V
        to a unit x *without knowledge of the actual V-to-x conversion*. Thus,
        with the exception of the case of the topography channel, this is a
        'pseudo-unit'.
    - V_to_x_conversion is a conversion factor from this 'pseudo-unit' to
        physical units. Another way to think of this is that the 'pseudo-unit'
        considers V_to_x_conversion to be 1-to-1, and this is a correction of
        that (for the cases where the conversion is *not* 1-to-1).

    Args:
        ds: the Dataset instance we are to convert, assumed to be from a gxsm
            data file.
        channel_config: a GxsmChannelConfig instance, holding the necessary
            info to convert and name our new data variable.

    Returns:
        A modified Dataset, where the 'FloatField' variable has been replaced
        by a variabled named after its channel, with physical units stored.

        Note: the channel name will either be taken directly from the gxsm file
        save format (e.g. topography files are saved with "Topo"), *or* with
        the desired channel name indicated in the config.

    Raises:
        None.
    """
    converted_data = (ds[GXSM_DATA_VAR] * ds[GXSM_DATA_DIFFERENTIAL]
                      * channel_config.conversion_factor)

    converted_data.attrs['units'] = channel_config.units
    ds[channel_config.name] = converted_data

    # Delete the original data variables
    ds = ds.drop_vars(GXSM_KEPT_DATA_VARS)
    return ds


def clean_kept_coords(ds: xarray.Dataset) -> xarray.Dataset:
    """Update kept coords to be in nm and have 'units' attr.

    Helper to ensure our spatial coordinate dimensions (GXSM_KEPT_DATA_COORDS)
    are properly labeled and using SI units.

    Args:
        ds: xarray Dataset instance.

    Returns:
        xarray.Dataset instance, with GXSM_KEPT_DATA_COORDS converted to nm
        and with a 'units' attr indicating as such.
    """
    for coord in GXSM_KEPT_COORDS:
        # Convert kept coords to 'nm' and add metadata units
        updated_coord = ds[coord] * cc.GXSM_TOPO_CONVERSION_FACTOR
        updated_coord.attrs['units'] = cc.GXSM_TOPO_UNITS
        ds[coord] = updated_coord
    return ds


def clean_up_metadata(ds: xarray.Dataset, saved_vars_list: list = []
                      ) -> xarray.Dataset:
    """Convert 'metadata' variables to attributes.

    gxsm does not store *all* metadata as attributes in its NetCDF file.
    Instead, many of them are stored as variables. This is to better qualify
    them: a description (long_name) and units (var_unit) attributes are
    provided.

    However, we find this introduces confusion between metadata and actual data
    (note that there are >100 metadata variables!). To better divide data and
    metadata, we provide this method; it moves all metadata variables to
    attributes, and places them appropriately within the xarray.Dataset.

    This means that the units are removed! Since this appears standard in
    other microscopy files, we believe it to be ok. A user curious about the
    units is encouraged to open the file using xarray directly, or study any
    appropriate documentation.

    Note that that we do not check vars in provided saved_vars_list for
    'units' attrs. We assume this will be (or has been) done separately.

    Args:
        ds: the Dataset instance we are to convert, assumed to be from a gxsm
            data file.
        saved_vars_list: a list of dataset variables to keep (i.e. not turn
            into attributes). This could contain any variables the user knows
            are not metadata.
        config: a dict containing gxsm channel configuration data (including
            the V-to-x unit conversion for each channel).

    Returns:
        A modified Dataset, where all metadata is stored as attributes.

    Raises:
        None.
    """
    # Create whitelist of hard-coded gxsm data vars to skip, as well as
    # desired channel names (in case we run this before or after
    # sanitizing the actual data).
    data_vars_whitelist = GXSM_KEPT_DATA_VARS + saved_vars_list

    # We are creating a new ds from grabbed data (rather than manually
    # removing vars/coords), since the latter threw strange exceptions.
    kept_data_vars = {}
    kept_coords = {}
    new_attrs = {}

    for coord in ds.coords:
        if coord not in GXSM_KEPT_COORDS:
            new_attrs[coord] = utils.extract_numpy_data(ds[coord].values)
        else:
            kept_coords[coord] = ds[coord]

    for var in ds.data_vars:
        if var not in data_vars_whitelist:
            new_attrs[var] = utils.extract_numpy_data(ds[var].values)
        else:
            kept_data_vars[var] = ds[var]

    full_attrs = {**ds.attrs, **new_attrs}
    new_ds = xarray.Dataset(data_vars=kept_data_vars, coords=kept_coords,
                            attrs=full_attrs)
    return new_ds


def is_gxsm_file(ds: xarray.Dataset) -> bool:
    """Check if the provided file is a supported gxsm file."""
    try:
        return GXSM_FORMAT_CHECK[1] in ds.attrs[GXSM_FORMAT_CHECK[0]].lower()
    except KeyError:
        return False



"""Logic around reading spectroscopy files."""

import numpy as np


VISUALIZATION_HEADER = '# view'
COMMENT = '#C'

MD_STRIP_CHARS = '# '
MD_KEY_VAL_DELINEATOR = '::'

DATA_START = '#C Data Table'
DATA_END = '#C END.'
DATA_SEP = '\t'
UNIT_STRIP_CHARS = r'(")'

# Extracting useful metadata
SUBKEY_SEP = '='
POS_STRIP_CHARS = ' ,'
IN_KEYS_FILENAME = ['FileName', 'name']
IN_KEYS_DATE = ['Date', 'date']
IN_KEYS_PROBE_POS = ['GXSM-Main-Offset', ['X0', 'Y0']]

KEY_PROBE_POS_X = 'probe_position_x'
KEY_PROBE_POS_Y = 'probe_position_y'
KEY_PROBE_POS_UNITS = 'probe_position_units'
KEY_FILENAME = 'filename'
KEY_DATE = 'date'
KEY_UNITS = 'units'


def validate_spec_file(lines: list[str]):
    """Run quick checks to make sure this is a spec file."""
    assert lines[0].startswith(VISUALIZATION_HEADER)


def extract_raw_metadata(lines: list[str]) -> dict[str, str]:
    """Extract metadata from spec file lines into dict.

    Given the read lines from a spectroscopy file, extract the raw metadata
    and return it as a dict of metadata key and vals (both str).

    This metadata is of format:
        # KEY :: SUB_KEY1=... SUB_KEY2=...

    This method will extract the keys as independent attributes, with the full
    val string (consisting of various sub keys) stored as a metadata element.

    Args:
        lines: list[str] of read lines from a spectroscpy file.

    Returns:
        str:str key:val dict containing METADATA_KEY:METADATA_STR.
    """
    comment_indices = [i for i, v in enumerate(lines) if
                       v.startswith(COMMENT)]

    # Extract the metadata (OWN METHOD)
    raw_metadata = {}
    md_lines = lines[1:comment_indices[0]]

    # Skip lines without metadata key:vals
    md_lines = [line for line in md_lines if MD_KEY_VAL_DELINEATOR in line]

    for line in md_lines:
        kv = line.split(MD_KEY_VAL_DELINEATOR)
        k = kv[0].strip(MD_STRIP_CHARS)
        v = kv[1].strip()
        raw_metadata[k] = v

    return raw_metadata


def parse_useful_metadata(raw_metadata: dict[str, str]) -> dict[str, str]:
    """Parse raw metadata and extract some useful metadata.

    Particularly, we want to explicit:
    - Probe Position (x,y) where the spec file was collected, stored with
        KEY_PROBE_POS_X, KEY_PROBE_POS_Y, KEY_PROBE_POS_UNITS
    - Filename of the spec file, stored with KEY_FILENAME.
    - Date of the collection, stored with KEY_DATE.

    Args:
        raw_metadata: raw metadata extracted from spec file.

    Returns:
        new dict containing the parsed 'useful' data, with keys matching
            those indicated above.
    """
    useful_metadata = {}
    useful_metadata[KEY_DATE] = _extract_date(raw_metadata)
    useful_metadata[KEY_FILENAME] = _extract_filename(raw_metadata)

    coords, units = _extract_position(raw_metadata)
    useful_metadata[KEY_PROBE_POS_X] = coords[0]
    useful_metadata[KEY_PROBE_POS_Y] = coords[1]
    useful_metadata[KEY_PROBE_POS_UNITS] = units[0]
    return useful_metadata


def _extract_date(raw_metadata: dict[str, str]) -> str:
    date_subdata = raw_metadata[IN_KEYS_DATE[0]].split(SUBKEY_SEP)
    date_idx = [i for i, v in enumerate(date_subdata) if
                IN_KEYS_DATE[1] in v]
    assert len(date_idx) == 1
    date_idx = date_idx[0]
    return date_subdata[date_idx + 1]


def _extract_position(raw_metadata: dict[str, str]) -> (list[str], list[str]):
    pos_subdata = raw_metadata[IN_KEYS_PROBE_POS[0]].split(SUBKEY_SEP)

    coords = []
    units = []
    for pos in range(2):
        idx = [i for i, v in enumerate(pos_subdata) if
               IN_KEYS_PROBE_POS[1][pos] in v]
        idx = idx[0]

        # Expect: '0 Ang  ' or '0 Ang, '
        subdata = pos_subdata[idx + 1].split()
        coords.append(subdata[0].strip(POS_STRIP_CHARS))
        units.append(subdata[1].strip(POS_STRIP_CHARS))

    assert units[0] == units[1]
    return coords, units


def _extract_filename(raw_metadata: dict[str, str]) -> str:
    filename_subdata = raw_metadata[IN_KEYS_FILENAME[0]].split(SUBKEY_SEP)
    filename_idx = [i for i, v in enumerate(filename_subdata) if
                    IN_KEYS_FILENAME[1] in v]
    assert len(filename_idx) == 1
    filename_idx = filename_idx[0]
    return filename_subdata[filename_idx + 1]


def extract_data(lines: list[str]) -> (list[str], list[str], np.ndarray):
    r"""Extract data from spec file lines into names, units, and data.

    Given the read lines from a spectroscopy file, extract the data (as a
    numpy array), the channel names (as a list of strings), and the units
    (as a list of strings).

    The data if stored in this format:
        #C Data Table             :: data=
        #C Index\t"ADC0-I (nA)"\t [...]
        0\t1.1111\t2.222\t [...]
        [...]
        #C
        #C END.

    We use the first and last line expectations to constraint the data.
    We extract the units and channel names from the 2nd line.
    We extract all data between the 2nd line and the end.

    Args:
        lines: list[str] of read lines from a spectroscpy file.

    Returns:
        (list[str], list[str], np.ndarray) of:
        - list[str] of channel names
        - list[str] of units (empty string for Index/Block-start-index)
        - np.ndarray of all data.
    """
    data_start_indices = [i for i, v in enumerate(lines) if
                          v.startswith(DATA_START)]
    data_end_indices = [i for i, v in enumerate(lines) if
                        v.startswith(DATA_END)]
    data_range = (data_start_indices[0], data_end_indices[0])

    # Get data header
    data_header = lines[data_range[0] + 1].split(DATA_SEP)
    names, units = _extract_names_units(data_header)

    # Get data
    data_lines = lines[data_range[0] + 2:data_range[1] - 1]
    data = _extract_data(data_lines, len(names))

    return names, units, data


def _extract_names_units(data_header: str) -> (list[str], list[str]):
    # Remove '#C ' from first item (index)
    data_header[0] = data_header[0][3::]

    channel_names = [substr.split()[0].strip(UNIT_STRIP_CHARS)
                     for substr in data_header]
    units = [substr.split()[1].strip(UNIT_STRIP_CHARS)
             if UNIT_STRIP_CHARS[0] in substr else ''
             for substr in data_header]
    assert units[0] == ''  # Index should have no units

    return channel_names, units


def _extract_data(data_lines: list[str], num_cols: int) -> np.ndarray:
    data = [line.split(DATA_SEP) for line in data_lines]
    for line in data:
        assert num_cols == len(line)
    data = np.array(data, np.float32)
    return data



"""Channel Config File Access / Explanation.

This file holds the channel-config file logic: a method to read the file,
and a class to configure a channel for conversion. All of the 'options'
around converting to physical units is within the class's constructor.

The config file is a TOML file, which is read easily enough. Look in the
examples directory for a sample.

The options logic (and how the config file should be written) can be read
from the class documentation.
"""

from dataclasses import dataclass
import tomli
import xarray
from . import utils
from . import filename as fn

# Hard-coded conversion info that may exist in the metadata
# TODO: investigate if other params are stored (e.g. dHertz2V)
# TODO: Validate whether ADC0mITunnel is the only way ADC0 is saved!?!?!
GXSM_CHANNEL_METADATA_DICT = {'ADC0mITunnel':
                              {'name': 'sranger_mk2_hwi_XSM_Inst_nAmpere2V',
                               'units': 'nA'}}

# Special case! We treat the topography in a unique way.
GXSM_CHANNEL_TOPO = 'Topo'
GXSM_TOPO_UNITS = 'nm'  # gxsm data stored in angstrom, converted to nm here
GXSM_TOPO_CONVERSION_FACTOR = 0.1

DEFAULT_UNITS = 'raw'
DEFAULT_CONVERSION_FACTOR = 1.0


@dataclass
class GxsmChannelConfig:
    """Class holds the attribs to convert to 'physical' units.

    You could argue this is overkill: tomli.load() will already give us a
    multi-level dict! We mainly add this for 2 reasons:
        1. Encapsulate the config logic from elsewhere. We now have the
            'config' logic handled within here.
        2. Document what is expected from each channel in the .toml.

    Below, we explain how the toml file should be written (and the associated
    attributes). We use toml_dict to refer to the dict returned from reading
    the toml file.

    So: each 'channel' should look like this in the .toml:
    [$gxsm_channel_name$-$scan_direction$]
    name: ...
    conversion_factor: ...
    units: ...

    where, $gxsm_channel_name$ is the name of the channel as saved by gxsm
    (see GxsmFileAttribs descriptoin for more info).
    E.g. "Topo" for topography, "ADC0" for ADC0.

    Notes:
        - When we create the name for the *stored* variable, we must
            distinguish between scan directions (because we may have saved
            both directions on one channel). Thus, the saved data var will have
            a name that is:
                config.name + config.scan_direction.
        - Topography is a special case! It is somewhat 'handled' by gxsm and
            stored in Angstrom. We convert to nm here. However, we *do not*
            inhibit you from doing any conversion on your end. If you *do*
            override this behavior, make sure you consider that the base
            unit provided by gxsm is in Angstrom, and convert accordingly!

    Attributes:
        name: the name we should provide this channel in the final
            xarray.Dataset. If not in the toml_dict, we use $gxsm_channel_name$.
        conversion_factor: the factor to apply to the 'raw' units, to convert
            them to real units. Used in preprocess.py. If not in the toml_dict,
            we (optionally) first check if this data can be found in the
            metadata; if not, we raise ane exception.
        units: the true units we are saving in (if saving physical units). For
            the final created variable, we will add a 'units' attrib with this
            string. If not in the toml_dict, we (optoinally) first check if
            the data can be found in the metadata; if not, we raise an exception.
    """

    name: str
    conversion_factor: float
    units: str


def CreateGxsmChannelConfig(channels_config_dict: dict | None,
                            ds: xarray.Dataset | None,
                            file_attribs: fn.GxsmFileAttribs,
                            use_physical_units: bool,
                            allow_convert_from_metadata: bool
                            ) -> GxsmChannelConfig:
    """Create GxsmChannelConfig from input data.

    This contains the complicated 'config' setup logic. Read carefully!

    Args:
        channels_config_dict: a dict containing gxsm channel configuration
            data (some subset of these attributes) for each channel of
            interest.
        ds: the Dataset instance associated with this channel, used to
            extract metadata (if necessary). If None, we do not try to
            use it when units are not provided (and physical units are
            requested).
        file_attribs: Gxsm file attributes, for easy access to 'channel'
            name.
        use_physical_units: whether or not to record the data in physical
            units. If true, we require 'conversion_factor'  and 'units'
            to exist (see optional exception below). Else, 'units' will be
            'raw' and 'conversion_factor' '1.0'.
        allow_convert_from_metadata: for some channels, there are hardcoded
            gxsm metadata attributes that contain the V-to-units conversion.
            If this attribute is true and toml_dict *does not contain* a
            conversion_factor, we will try to find a suitable conversion
            from the metadata.

    Returns:
        A GxsmChannelConfig instance.

    Raises:
        - KeyError if a key is missing from the channel_config_dict (or,
            if allowing metadata, the metadata attribs)
        - ValueError if the conversion factor provided / read is 0!
    """
    d = channels_config_dict
    c = file_attribs.channel
    MAP = GXSM_CHANNEL_METADATA_DICT

    # Recall the saved data var name must be unique (can have 2 scans
    # per channel)
    try:
        name = d[c]['name']
    except (KeyError, TypeError):
        name = c
    name = fn.get_unique_channel_name(name, file_attribs.scan_direction)

    # We allow the user to overide topography, but they must input
    # all parameters themselves
    if c == GXSM_CHANNEL_TOPO and (not d or GXSM_CHANNEL_TOPO not in d):
        conversion_factor = GXSM_TOPO_CONVERSION_FACTOR
        units = GXSM_TOPO_UNITS
    elif use_physical_units:
        # If no dict was passed, create an empty one (for code simplicity)
        d = {} if not d else d
        try:
            conversion_factor = d[c]['conversion_factor']
            units = d[c]['units']
        except (KeyError) as e:
            # Because of this, we *must* run this on the original ds
            # (we are grabbing metadata as data var, where we will be
            # turning it into an attr).
            if (allow_convert_from_metadata and c in MAP and
               ds and MAP[c]['name'] in ds.data_vars):
                conversion_factor = utils.extract_numpy_data(
                    ds[MAP[c]['name']].values)
                units = MAP[c]['units']
                if conversion_factor == 0.0:
                    raise ValueError('Conversion factor for {} found in \
                    metadata, but was 0. Cannot continue.')
            else:
                raise type(e)(f'No conversion factor found for {c} (or '
                              'in metadata if permitted).') from e
    else:
        conversion_factor = DEFAULT_CONVERSION_FACTOR
        units = DEFAULT_UNITS
    return GxsmChannelConfig(name, conversion_factor, units)


def load_channels_config_dict(filename: str | None) -> dict:
    """Load config dict (currently a toml file)."""
    if filename:
        with open(filename, 'rb') as f:
            return tomli.load(f)
    else:
        return None



"""Contains read methods for a gxsm file.

This file contains the main methods to read from gxsm files and
convert to a NETCDF4 / HDF5 file.

For simplicity, we simply override the two main dataset accessor files
from xarray: open_dataset() and open_mfdataset() (allowing usage of
all the nifty xarray features, like dask).

Read the description of each method for more info.

Basic usage:
    ds = open_dataset(...)
    multifile_ds = open_mfdataset(...)
"""

import xarray
import pandas as pd
from pathlib import Path
from functools import partial
from . import channel_config as cc
from . import preprocess as pp
from . import spec


def open_mfdataset(paths: str | list[str | Path],
                   channels_config_path: str | Path | None = None,
                   use_physical_units: bool = True,
                   allow_convert_from_metadata: bool = False,
                   simplify_metadata: bool = True, **kwargs
                   ) -> xarray.Dataset:
    """Open multiple files as a single dataset.

    Wrapper on top of xarray.open_mfdataset(), for handling gxsm netcdf
    files. It opens the different .nc files as a single xarray.Dataset
    instance, expecting them to each correspond to a channel of the same
    collection.

    Note that gxsm .nc files contain the raw DAC counter data for each
    channel, rather than physical units. Additionally, many metadata
    attributes are saved as variables rather than attributes (see preprocess.py
    for further explanation). The added options allow the user to choose to
    'fix' these differences.

    Lastly, in order to merge the data, we have to hard-code the following
    open_mfdataset params:
        - concat_dim=None: i.e., no concatenation, since each is its own
            channel.
        - combine='nested': to allow only merging.
        - compat='identical': to enforce that we expect all same-named
            variables to be identical.

    Args:
        paths: either a string glob in the form "path/to/my/files/*.nc" or an
            explicit list of files to open. Paths can be given as strings or
            as pathlib Paths. Note that the xarray option of nested
            list-of-lists is not supported here.
        channels_config_path: either a string or Path to a 'channels config'
            toml file, containing the channel information for raw-to-physical
            conversion. See examples/channels_config.toml for an example.
        use_physical_units: whether or not to record the data in physical
            units. If true, we require 'conversion_factor'  and 'units'
            to exist (see optional exception below). Else, 'units' will be
            'raw' and 'conversion_factor' '1.0'.
        allow_convert_from_metadata: for some channels, there are hardcoded
            gxsm metadata attributes that contain the V-to-units conversion.
            If this attribute is true, we will use the metadata conversion
            as a fallback (i.e. if the config does not contain it).
        simplify_metadata: whether or not to convert all metadata variables
            to attributes.

    Returns:
        An xarray.Dataset instance, with each file's data being stored as a data
        variable named $channel$, where $channel$ is the channel substring
        in the filepath plus the scan direction.
    """
    # Getting open_mfdataset() args from kwargs if user provided (and setting
    # our known default, for it to work). Basically, we trust the user to
    # know what they are doing.
    combine = kwargs.get('combine', 'nested')
    compat = kwargs.get('compat', 'identical')
    join = kwargs.get('join', 'outer')

    channels_config_dict = cc.load_channels_config_dict(channels_config_path)
    partial_func = partial(pp.preprocess,
                           use_physical_units=use_physical_units,
                           allow_convert_from_metadata=allow_convert_from_metadata,
                           simplify_metadata=simplify_metadata,
                           channels_config_dict=channels_config_dict)
    # Note: in principle, we could use combine='by_coords'. For some reason,
    # it appears that using this (instead of 'nested') causes the combination
    # of attributes (metadata) to miss some (presumably, because they only
    # exist in the main file). For now, sticking with nested. In the future,
    # we could try both.
    return xarray.open_mfdataset(paths, preprocess=partial_func,
                                 concat_dim=None, combine=combine,
                                 compat=compat, join=join, **kwargs)


def open_dataset(filename_or_obj: str | Path,
                 channels_config_path: str | Path | None = None,
                 use_physical_units: bool = True,
                 allow_convert_from_metadata: bool = False,
                 simplify_metadata: bool = True, **kwargs
                 ) -> xarray.Dataset:
    """Open and decode a dataset from a file or file object.

    Wrapper on top of xarray.open_dataset(), for handling gxsm nc
    files. It opens an .nc file as a single xarray.Dataset
    instance, with the 'z-data' stored as a data variable with name
    $channel$, where $channel$ is the channel name for the file.

    Note that gxsm .nc files contain the raw DAC counter data for each
    channel, rather than phyiscal units. Additionally, many metadata
    attributes are saved as variables rather than attributes (see preprocess.py
    for further explanation). The added options allow the user to choose to
    'fix' these differences.

    Args:
        filename_or_obj: path of the file to open, can be given as a string
            or a pathlib Path. Note that other xarray type options are not
            supported here.
        channels_config_path: either a string or Path to a 'channels config'
            toml file, containing the channel information for raw-to-physical
            conversion. See examples/channels_config.toml for an example.
        use_physical_units: whether or not to record the data in physical
            units. If true, we require 'conversion_factor'  and 'units'
            to exist (see optional exception below). Else, 'units' will be
            'raw' and 'conversion_factor' '1.0'.
        allow_convert_from_metadata: for some channels, there are hardcoded
            gxsm metadata attributes that contain the V-to-units conversion.
            If this attribute is true, we will use the metadata conversion
            as a fallback (i.e. if the config does not contain it).
        simplify_metadata: whether or not to convert all metadata variables
            to attributes.

    Returns:
        An xarray.Dataset instance, with the file's data being stored as a data
        variable named $channel$, where $channel$ is the channel name for the
        file in the file structure.
    """
    channels_config_dict = cc.load_channels_config_dict(channels_config_path)
    ds = xarray.open_dataset(filename_or_obj, **kwargs)
    return pp.preprocess(ds, use_physical_units=use_physical_units,
                         allow_convert_from_metadata=allow_convert_from_metadata,
                         simplify_metadata=simplify_metadata,
                         channels_config_dict=channels_config_dict)


def open_spec(filename: str | Path) -> pd.DataFrame:
    """Open and decode spec data from a file or file object.

    Reads gxsm 'VP' style spectroscopy files (Vector Probe), converting them to
    a DataFrame. These spectroscopy files are tabular data, meaning we have
    N data rows of M columns (corresponding to various channels recorded),
    and thus an NxM array.

    A file also has a 'header' of various metadata attributes, written in a
    key/subkey format:
    # KEY :: SUB_KEY1=... SUB_KEY2=...

    This method will extract the keys as independent attributes, with the full
    val string (consisting of various sub keys) stored as a metadata element.
    However, we make an exception for 4 metadata attributes we deem  worth
    expliciting: x and y coordinates where the spectroscopy was collected,
    the filename, and the data of the collection.

    Args:
        filename: path of the file to open, can be given as a string
            or a pathlib Path.

    Returns:
        A pandas.DataFrame instance, with the file's data stored with channels
        explicited and units as an attr in each Series' attr instance.
    """
    raw_metadata = {}
    data = None
    names = None
    units = None
    with open(filename, 'r') as file:
        lines = file.readlines()
        spec.validate_spec_file(lines)
        raw_metadata = spec.extract_raw_metadata(lines)
        names, units, data = spec.extract_data(lines)

    new_metadata = spec.parse_useful_metadata(raw_metadata)
    metadata = raw_metadata | new_metadata

    df = pd.DataFrame(data, columns=names)
    for k, v in metadata.items():
        df.attrs[k] = v

    # Add units with UNITS_KEY as dict to attributes.
    units_dict = dict(zip(names, units))
    df.attrs[spec.KEY_UNITS] = units_dict

    return df
